""" Administrative access to Pin data """
from django.contrib import admin
from django.utils.translation import ugettext_lazy as _

from pinpayments.models import (
    PinRecipient, PinTransfer, PinTransaction, CustomerToken, PinPlan, Subscription
)


class PinTransactionAdmin(admin.ModelAdmin):
    """ Inspect transactions from here """
    list_display = (
        'date',
        'processed',
        'succeeded',
        'environment',
        'amount',
        'currency',
        'email_address',
        'pin_response',
        'ip_address',
        'transaction_token',
    )
    search_fields = (
        'email_address',
        'ip_address',
        'card_token',
        'transaction_token',
    )
    list_filter = ('processed', 'succeeded', 'environment', 'currency')
    date_hierarchy = 'date'
    readonly_fields = (
        'date',
        'description',
        'processed',
        'succeeded',
        'environment',
        'amount',
        'currency',
        'fees',
        'email_address',
        'pin_response',
        'ip_address',
        'transaction_token',
        'card_token',
        'customer_token',
        'email_address',
        'card_address1',
        'card_address2',
        'card_city',
        'card_state',
        'card_postcode',
        'card_country',
        'card_number',
        'card_type',
        'pin_response_text',
    )


class PinTransactionInline(admin.TabularInline):
    """
    Used to show transactions for a particular customer token, if using
    the Customer API.
    """
    model = PinTransaction
    fields = (
        'date',
        'processed',
        'succeeded',
        'environment',
        'amount',
        'currency',
        'transaction_token',
        'ip_address',
    )
    readonly_fields = fields
    extra = 0
    can_delete = False

    def has_add_permission(self, request):
        return False


class TokenAdmin(admin.ModelAdmin):
    """ Shows customer tokens """
    list_display = (
        'token',
        'user',
        'environment',
        'created',
        'active',
        'card_type',
        'card_number',
    )
    search_fields = ('token', 'card_number')
    list_filter = ('environment', 'card_type', 'active')
    date_hierarchy = 'created'
    inlines = (PinTransactionInline,)
    readonly_fields = ('environment', 'token', 'card_type', 'card_number')


class PinTransferAdmin(admin.ModelAdmin):
    """ Shows the details of a transfer """
    list_display = (
        'created',
        'get_value',
        'recipient',
        'status',
    )
    search_fields = (
        'recipient',
        'transfer_token',
    )
    date_hierarchy = 'created'
    readonly_fields = (
        'transfer_token',
        'status',
        'currency',
        'description',
        'amount',
        'recipient',
        'created',
        'pin_response_text',
    )

    def has_add_permission(self, request):
        return False

    def get_value(self, obj):
        return "{0:.2f} {1}".format(obj.value, obj.currency)
    get_value.short_description = _('Value')
    get_value.admin_order_field = 'amount'


class PinTransferInline(admin.TabularInline):
    """ Shows transfers under recipients """
    model = PinTransfer
    fields = [
        'created',
        'get_value',
        'status',
        'currency',
        'description',
        'transfer_token',
    ]
    readonly_fields = fields
    extra = 0
    can_delete = False

    def has_add_permission(self, request):
        return False

    def get_value(self, obj):
        return "{0:.2f} {1}".format(obj.value, obj.currency)
    get_value.short_description = _('Value')


class PinRecipientAdmin(admin.ModelAdmin):
    """ Allows viewing and re-aliasing PinRecipients """
    list_display = (
        'token',
        'email',
        'name',
        'environment',
        'created',
        'bank_account',
    )
    search_fields = ('token', 'email', 'name')
    list_filter = ('environment',)
    date_hierarchy = 'created'
    inlines = (PinTransferInline,)
    readonly_fields = list_display  # all the fields

class PinPlanAdmin(admin.ModelAdmin):
    """ Allows viewing payment plans. They should be created on the Pin dashboard. """
    list_display = (
        'plan_token',
        'name',
        'environment',
        'amount',
        'currency',
        'created',
    )
    search_fields = ('token', 'name',)
    list_filter = ('environment', 'currency',)
    date_hierarchy = 'created'

    def has_add_permission(self, request):
        return False

admin.site.register(PinRecipient, PinRecipientAdmin)
admin.site.register(PinTransaction, PinTransactionAdmin)
admin.site.register(CustomerToken, TokenAdmin)
admin.site.register(PinTransfer, PinTransferAdmin)
admin.site.register(PinPlan, PinPlanAdmin)
admin.site.register([Subscription])



# Generated by Django 3.0.5 on 2020-04-23 02:53

from decimal import Decimal
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='BankAccount',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('token', models.CharField(db_index=True, help_text='A bank account token provided by Pin', max_length=40, verbose_name='Pin API Bank account token')),
                ('bank_name', models.CharField(help_text='The name of the bank at which this account is held', max_length=100, verbose_name='Bank Name')),
                ('branch', models.CharField(blank=True, help_text='The name of the branch at which this account is held', max_length=100, verbose_name='Branch name')),
                ('name', models.CharField(help_text='The name of the bank account', max_length=100, verbose_name='Recipient Name')),
                ('bsb', models.IntegerField(help_text='The BSB (Bank State Branch) code of the bank account.', verbose_name='BSB')),
                ('number', models.CharField(help_text='The account number of the bank account', max_length=20, verbose_name='BSB')),
                ('environment', models.CharField(blank=True, db_index=True, help_text='The name of the Pin environment to use, eg test or live.', max_length=25)),
            ],
        ),
        migrations.CreateModel(
            name='CustomerToken',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('environment', models.CharField(blank=True, db_index=True, help_text='The name of the Pin environment to use, eg test or live.', max_length=25)),
                ('token', models.CharField(help_text='Generated by Card API or Customers API', max_length=100, verbose_name='Token')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='Created')),
                ('active', models.BooleanField(default=True, verbose_name='Active')),
                ('card_type', models.CharField(blank=True, choices=[('master', 'Mastercard'), ('visa', 'Visa')], help_text='Determined automatically by Pin', max_length=20, null=True, verbose_name='Card Type')),
                ('card_number', models.CharField(blank=True, help_text='Cleansed by Pin API', max_length=100, null=True, verbose_name='Card Number')),
                ('card_name', models.CharField(blank=True, max_length=100, null=True, verbose_name='Name on Card')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='PinRecipient',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('token', models.CharField(db_index=True, help_text='A recipient token provided by Pin', max_length=40)),
                ('email', models.EmailField(help_text='As passed to Pin.', max_length=100)),
                ('name', models.CharField(blank=True, help_text='Optional. The name by which the recipient is referenced', max_length=100, null=True)),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='Time created')),
                ('environment', models.CharField(blank=True, db_index=True, help_text='The name of the Pin environment to use, eg test or live.', max_length=25)),
                ('bank_account', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='pinpayments.BankAccount')),
            ],
        ),
        migrations.CreateModel(
            name='PinTransfer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('transfer_token', models.CharField(blank=True, db_index=True, help_text='Unique ID from Pin for this transfer', max_length=100, null=True, verbose_name='Pin API Transfer Token')),
                ('status', models.CharField(blank=True, help_text='Status of transfer at time of saving', max_length=100, null=True)),
                ('currency', models.CharField(help_text='currency of transfer', max_length=10)),
                ('description', models.CharField(blank=True, help_text='Description as shown on statement', max_length=100, null=True)),
                ('amount', models.IntegerField(help_text='Transfer amount, in the base unit of the currency (e.g.: cents for AUD, yen for JPY)')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('pin_response_text', models.TextField(blank=True, help_text='The full JSON response from the Pin API', null=True, verbose_name='Complete API Response')),
                ('recipient', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='pinpayments.PinRecipient')),
            ],
        ),
        migrations.CreateModel(
            name='PinTransaction',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(db_index=True, help_text='Time this transaction was put in the database. May differ from the time that PIN reports the transaction.', verbose_name='Date')),
                ('environment', models.CharField(blank=True, db_index=True, help_text='The name of the Pin environment to use, eg test or live.', max_length=25)),
                ('amount', models.DecimalField(decimal_places=2, max_digits=10, verbose_name='Amount (Dollars)')),
                ('fees', models.DecimalField(blank=True, decimal_places=2, default=Decimal('0.00'), help_text='Fees charged to you by Pin, for this transaction, in dollars', max_digits=10, null=True, verbose_name='Transaction Fees')),
                ('description', models.TextField(blank=True, help_text='As provided when you initiated the transaction', null=True, verbose_name='Description')),
                ('processed', models.BooleanField(default=False, help_text='Has this been sent to Pin yet?', verbose_name='Processed?')),
                ('succeeded', models.BooleanField(default=False, help_text='Was the transaction approved?', verbose_name='Success?')),
                ('currency', models.CharField(default='AUD', help_text='Currency transaction was processed in', max_length=100, verbose_name='Currency')),
                ('transaction_token', models.CharField(blank=True, db_index=True, help_text='Unique ID from Pin for this transaction', max_length=100, null=True, verbose_name='Pin API Transaction Token')),
                ('card_token', models.CharField(blank=True, help_text='Card token used for this transaction (Card API and Web Forms)', max_length=40, null=True, verbose_name='Pin API Card Token')),
                ('pin_response', models.CharField(blank=True, help_text='Response text, usually Success!', max_length=255, null=True, verbose_name='API Response')),
                ('ip_address', models.GenericIPAddressField(help_text='IP Address used for payment')),
                ('email_address', models.EmailField(help_text='As passed to Pin.', max_length=100, verbose_name='E-Mail Address')),
                ('card_address1', models.CharField(blank=True, help_text='Address entered by customer to process this transaction', max_length=100, null=True, verbose_name='Cardholder Street Address')),
                ('card_address2', models.CharField(blank=True, max_length=100, null=True, verbose_name='Cardholder Street Address Line 2')),
                ('card_city', models.CharField(blank=True, max_length=100, null=True, verbose_name='Cardholder City')),
                ('card_state', models.CharField(blank=True, max_length=100, null=True, verbose_name='Cardholder State')),
                ('card_postcode', models.CharField(blank=True, max_length=100, null=True, verbose_name='Cardholder Postal / ZIP Code')),
                ('card_country', models.CharField(blank=True, max_length=100, null=True, verbose_name='Cardholder Country')),
                ('card_number', models.CharField(blank=True, help_text='Cleansed by Pin API', max_length=100, null=True, verbose_name='Card Number')),
                ('card_type', models.CharField(blank=True, choices=[('master', 'Mastercard'), ('visa', 'Visa')], help_text='Determined automatically by Pin', max_length=20, null=True, verbose_name='Card Type')),
                ('pin_response_text', models.TextField(blank=True, help_text='The full JSON response from the Pin API', null=True, verbose_name='Complete API Response')),
                ('customer_token', models.ForeignKey(blank=True, help_text='Provided by Customer API', null=True, on_delete=django.db.models.deletion.SET_NULL, to='pinpayments.CustomerToken')),
            ],
            options={
                'verbose_name': 'PIN.net.au Transaction',
                'verbose_name_plural': 'PIN.net.au Transactions',
                'ordering': ['-date'],
            },
        ),
    ]



"""
Non-model related objects
"""

from decimal import Decimal

from django.conf import settings
import requests

from .exceptions import ConfigError, PinError


class PinEnvironment(object):
    """ Container for pin settings """
    def __init__(self, name="test", *args, **kwargs):
        """ Populate contents from Settings """
        if name in ('test', ''):
            name = getattr(settings, 'PIN_DEFAULT_ENVIRONMENT', 'test')

        env_dict = {}
        try:
            env_dict = settings.PIN_ENVIRONMENTS[name]
        except KeyError:
            raise ConfigError("Invalid environment {0}".format(name))

        if not set(env_dict.keys()).issuperset(['key', 'secret', 'host']):
            raise ConfigError(
                "The environment {0} was not correctly defined.\n"
                "Please provide the three fields:\n"
                "\t'key', 'secret', and 'host'".format(name)
            )

        self.name = name
        self.host = env_dict['host']
        self.key = env_dict['key']
        self.secret = env_dict['secret']
        super(PinEnvironment, self).__init__(*args, **kwargs)

    @property
    def auth(self):
        """ Returns auth as expected by requests for Pin """
        return (self.secret, '')

    def _pin_request(self, method, url_tail, payload=None, always_return=False):
        """
        Internal method to abstract common details of calls to Pin API
        """
        method = method.lower()
        if method not in ['get', 'post', 'put']:
            raise Exception(
                "Method for request '{0}' was invalid".format(method)
            )
        requests_method = getattr(requests, method)
        url = 'https://{0}/1{1}'.format(self.host, url_tail)
        if payload is not None:
            response = requests_method(
                url,
                auth=self.auth,
                params=payload,
                headers={'content_type': 'application/json'}
            )
        else:
            response = requests_method(
                url,
                auth=self.auth,
                headers={'content_type': 'application/json'}
            )
        try:
            response_json = response.json()
        except (AttributeError, ValueError):
            if always_return:
                response_json = None
            else:
                raise PinError(
                    "Error retrieving response for environment {0}"
                    "at url {1}".format(self.name, url)
                )

        if not always_return:
            if 'error' in response_json.keys():
                raise PinError(
                    'Error returned from Pin API: {0}:{1}'.format(
                        response_json['error'],
                        response_json['error_description']
                    )
                )

        return (response, response_json)

    def pin_get(self, url_tail, always_return=False):
        """
        Provide a relative URL to access the API for it via GET
        Include the leading /
        Returns a tuple of the response and the decoded JSON
        Provide always_return=True to handle all errors yourself
        """
        return self._pin_request('GET', url_tail, always_return)

    def pin_put(self, url_tail, payload, always_return=False):
        """
        Provide a relative URL to access the API for it via PUT
        Include the leading /
        Returns a tuple of the response and the decoded JSON
        Provide always_return=True to handle all errors yourself
        """
        return self._pin_request('PUT', url_tail, payload, always_return)

    def pin_post(self, url_tail, payload, always_return=False):
        """
        Provide a relative URL to access the API for it via POST
        Include the leading /
        Returns a tuple of the response and the decoded JSON
        Provide always_return=True to handle all errors yourself
        """
        return self._pin_request('POST', url_tail, payload, always_return)

    def get_balance(self, currency="AUD"):
        """
        Query Pin for the balance of a Pin account in the currency given
        Returns a tuple containing Decimals of available and pending balance
        """
        response, response_json = self.pin_get('/balance')
        response_json = response_json['response']

        if not set(response_json.keys()).issuperset(['available', 'pending']):
            raise PinError(
                "The response from Pin was invalid for environment {0}\n"
                "details: {1}".format(self.name, response.text)
            )

        available = set([
            bal['amount'] if bal['currency'] == currency else None
            for bal in response_json['available']
        ])
        if None in available:
            available.remove(None)
        if len(available) != 1:
            raise PinError(
                "Error retrieving available balance for currency {0} "
                "in environment {1}. Available currencies and values are: \n"
                "\t{2}".format(currency, self.name, response_json['available'])
            )
        available_balance = Decimal(list(available)[0])

        pending = set([
            bal['amount'] if bal['currency'] == currency else None
            for bal in response_json['pending']
        ])
        if None in pending:
            pending.remove(None)
        if len(pending) != 1:
            raise PinError(
                "Error retrieving pending balance for currency {0}"
                "in environment {1}. Available currencies and values are: \n"
                "\t{2}".format(currency, self.name, response_json['pending'])
            )
        pending_balance = Decimal(list(pending)[0])

        return (available_balance, pending_balance)

    def get_available_balance(self, currency="AUD"):
        return self.get_balance(currency)[0]

    def get_pending_balance(self, currency="AUD"):
        return self.get_balance(currency)[1]



# Generated by Django 3.0.5 on 2020-04-27 07:03

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('pinpayments', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='PinPlan',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('environment', models.CharField(blank=True, db_index=True, help_text='The name of the Pin environment to use, eg test or live.', max_length=25)),
                ('plan_token', models.CharField(blank=True, db_index=True, help_text='Unique ID from Pin for this plan', max_length=100, null=True, verbose_name='Pin API Plan Token')),
                ('active', models.BooleanField(blank=True, default=True, help_text='Is this Plan still visible on the Pin API?')),
                ('name', models.CharField(blank=True, help_text='Description as shown on statement', max_length=100, null=True)),
                ('currency', models.CharField(help_text='Currency of plan', max_length=10)),
                ('amount', models.IntegerField(help_text='Charge amount, in the base unit of the currency (e.g.: cents for AUD, yen for JPY)')),
                ('interval', models.IntegerField(help_text='The interval between subsequent charges, interpreted in units defined in interval_units')),
                ('interval_unit', models.CharField(choices=[('day', 'Day(s)'), ('week', 'Week(s)'), ('month', 'Month(s)'), ('year', 'Year(s)')], help_text='The unit of measure applied to the interval amount.', max_length=10)),
                ('intervals', models.IntegerField(blank=True, default=0, help_text='Number of intervals before a subscription is automatically cancelled. Default 0 (no limit)')),
                ('setup_amount', models.IntegerField(blank=True, default=0, help_text='Amount to charge (in the currency base unit, eg cents for AUD) at the start of the first full interval.')),
                ('trial_amount', models.IntegerField(blank=True, default=0, help_text='Amount the customer will be charged in the currency base unit upon initiating a trial of this plan.')),
                ('trial_interval', models.IntegerField(blank=True, default=0, help_text='The interval between the start of the trial period and beginning of the paid subscription proper.')),
                ('trial_interval_unit', models.CharField(choices=[('day', 'Day(s)'), ('week', 'Week(s)'), ('month', 'Month(s)'), ('year', 'Year(s)')], help_text='The unit of measure applied to the trial interval amount.', max_length=10)),
                ('permission_cancel', models.BooleanField(blank=True, default=True, help_text='Can the customer cancel the subscription themselves via a link in their receipts?')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('pin_response_text', models.TextField(blank=True, help_text='The full JSON response from the Pin API', null=True, verbose_name='Complete API Response')),
            ],
        ),
        migrations.CreateModel(
            name='Subscription',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('subscription_token', models.CharField(blank=True, db_index=True, help_text='Unique ID from Pin for this subscription', max_length=100, null=True, verbose_name='Pin API Subscription Token')),
                ('environment', models.CharField(blank=True, db_index=True, help_text='The name of the Pin environment to use, eg test or live.', max_length=25)),
                ('active', models.BooleanField(blank=True, default=True, help_text='Is this subscription still visible on the Pin API?')),
                ('next_billing_date', models.DateTimeField(blank=True, help_text='The next time this subscription will be charged', null=True)),
                ('active_interval_started_at', models.DateTimeField(blank=True, help_text='When did the current trial or billing period begin?', null=True)),
                ('active_interval_finishes_at', models.DateTimeField(blank=True, help_text='When will the current trial or billing period finish?', null=True)),
                ('cancelled_at', models.DateTimeField(blank=True, help_text='When was the subscription cancelled?', null=True)),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='pinpayments.CustomerToken')),
                ('plan', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='pinpayments.PinPlan')),
            ],
        ),
    ]



""" Ensure that the models work as intended """
import json
from django.conf import settings
from django.contrib.auth.models import User
from django.test import TestCase
from django.test.utils import override_settings
from mock import patch
from pinpayments.models import (
    ConfigError,
    CustomerToken,
    PinError,
    PinTransaction
)
from requests import Response

ENV_MISSING_SECRET = {
    'test': {
        'key': 'key1',
        'host': 'test-api.pin.net.au',
    },
}

ENV_MISSING_HOST = {
    'test': {
        'key': 'key1',
        'secret': 'secret1',
    },
}


class FakeResponse(Response):
    def __init__(self, status_code, content):
        super(FakeResponse, self).__init__()
        self.status_code = status_code
        self._content = content.encode('utf-8')


class CustomerTokenTests(TestCase):
    # Need to override the setting so we can delete it, not sure why.
    @override_settings(PIN_DEFAULT_ENVIRONMENT=None)
    def test_default_environment(self):
        """
        Unset PIN_DEFAULT_ENVIRONMENT to test that the environment defaults
        to 'test'.
        """
        del settings.PIN_DEFAULT_ENVIRONMENT
        token = CustomerToken()
        token.user = User.objects.create()
        token.environment = None
        token.save()
        self.assertEqual(token.environment, 'test')


class CreateFromCardTokenTests(TestCase):
    """ Test the creation of customer tokens from card tokens """
    def setUp(self):
        """ Common setup for methods """
        super(CreateFromCardTokenTests, self).setUp()
        self.user = User.objects.create()
        self.response_data = json.dumps({
            'response': {
                'token': '1234',
                'email': 'test@example.com',
                'created_at': '2012-06-22T06:27:33Z',
                'card': {
                    'token': '54321',
                    'display_number': 'XXXX-XXXX-XXXX-0000',
                    'scheme': 'master',
                    'expiry_month': 6,
                    'expiry_year': 2017,
                    'name': 'Roland Robot',
                    'address_line1': '42 Sevenoaks St',
                    'address_line2': None,
                    'address_city': 'Lathlain',
                    'address_postcode': '6454',
                    'address_state': 'WA',
                    'address_country': 'Australia',
                    'primary': None,
                }
            }
        })
        self.response_error = json.dumps({
            'error': 'invalid_resource',
            'error_description':
                'One or more parameters were missing or invalid.'
        })

    @patch('requests.post')
    def test_default_environment(self, mock_request):
        """ return a default environment """
        mock_request.return_value = FakeResponse(200, self.response_data)
        token = CustomerToken.create_from_card_token('1234', self.user)
        self.assertEqual(token.environment, 'test')

    @override_settings(PIN_ENVIRONMENTS={})
    @patch('requests.post')
    def test_valid_environment(self, mock_request):
        """ Check errors are raised with no environments """
        mock_request.return_value = FakeResponse(200, self.response_data)
        with self.assertRaises(ConfigError):
            CustomerToken.create_from_card_token(
                '1234', self.user, environment='test'
            )

    @override_settings(PIN_ENVIRONMENTS=ENV_MISSING_SECRET)
    @patch('requests.post')
    def test_secret_set(self, mock_request):
        """ Check errors are raised when the secret is not set """
        mock_request.return_value = FakeResponse(200, self.response_data)
        with self.assertRaises(ConfigError):
            CustomerToken.create_from_card_token(
                '1234', self.user, environment='test'
            )

    @override_settings(PIN_ENVIRONMENTS=ENV_MISSING_HOST)
    @patch('requests.post')
    def test_host_set(self, mock_request):
        """ Check errors are raised when the host is not set """
        mock_request.return_value = FakeResponse(200, self.response_data)
        with self.assertRaises(ConfigError):
            CustomerToken.create_from_card_token(
                '1234', self.user, environment='test'
            )

    @patch('requests.post')
    def test_response_not_json(self, mock_request):
        """ Validate non-json response """
        mock_request.return_value = FakeResponse(200, '')
        with self.assertRaises(PinError):
            CustomerToken.create_from_card_token(
                '1234', self.user, environment='test'
            )

    @patch('requests.post')
    def test_response_error(self, mock_request):
        """ Validate generic error response """
        mock_request.return_value = FakeResponse(200, self.response_error)
        with self.assertRaises(PinError):
            CustomerToken.create_from_card_token(
                '1234', self.user, environment='test'
            )

    @patch('requests.post')
    def test_response_success(self, mock_request):
        """ Validate successful response """
        mock_request.return_value = FakeResponse(200, self.response_data)
        customer = CustomerToken.create_from_card_token(
            '1234', self.user, environment='test'
        )
        self.assertIsInstance(customer, CustomerToken)
        self.assertEqual(customer.user, self.user)
        self.assertEqual(customer.token, '1234')
        self.assertEqual(customer.environment, 'test')
        self.assertEqual(customer.card_number, 'XXXX-XXXX-XXXX-0000')
        self.assertEqual(customer.card_type, 'master')


class PinTransactionTests(TestCase):
    """ Transaction construction/init related tests """
    def setUp(self):
        """ Common setup for methods """
        super(PinTransactionTests, self).setUp()
        self.transaction = PinTransaction()
        self.transaction.card_token = '12345'
        self.transaction.ip_address = '127.0.0.1'
        self.transaction.amount = 500
        self.transaction.currency = 'AUD'
        self.transaction.email_address = 'test@example.com'
        self.transaction.environment = 'test'

    # Need to override the setting so we can delete it, not sure why.
    @override_settings(PIN_DEFAULT_ENVIRONMENT=None)
    def test_save_defaults(self):
        """
        Unset PIN_DEFAULT_ENVIRONMENT to test that the environment defaults
        to 'test'.
        """
        del settings.PIN_DEFAULT_ENVIRONMENT
        self.transaction.environment = None
        self.transaction.save()
        self.assertEqual(self.transaction.environment, 'test')
        self.assertTrue(self.transaction.date)

    def test_save_notokens(self):
        """
        Check that an error is thrown if neither card nor customer token
        are provided to the transaction
        """
        self.transaction.card_token = None
        self.transaction.customer_token = None
        self.assertRaises(PinError, self.transaction.save)

    def test_valid_environment(self):
        """
        Check that errors are thrown when a fake environment is requested
        """
        self.transaction.environment = 'this should not exist'
        self.assertRaises(PinError, self.transaction.save)


class ProcessTransactionsTests(TestCase):
    """ Transaction processing related tests """
    def setUp(self):
        """ Common setup for methods """
        super(ProcessTransactionsTests, self).setUp()
        self.transaction = PinTransaction()
        self.transaction.card_token = '12345'
        self.transaction.ip_address = '127.0.0.1'
        self.transaction.amount = 500
        self.transaction.currency = 'AUD'
        self.transaction.email_address = 'test@example.com'
        self.transaction.environment = 'test'
        self.transaction.save()
        self.response_data = json.dumps({
            'response': {
                'token': '12345',
                'success': True,
                'amount': 500,
                'total_fees': 500,
                'currency': 'AUD',
                'description': 'test charge',
                'email': 'test@example.com',
                'ip_address': '127.0.0.1',
                'created_at': '2012-06-20T03:10:49Z',
                'status_message': 'Success!',
                'error_message': None,
                'card': {
                    'token': 'card_nytGw7koRg23EEp9NTmz9w',
                    'display_number': 'XXXX-XXXX-XXXX-0000',
                    'scheme': 'master',
                    'expiry_month': 6,
                    'expiry_year': 2017,
                    'name': 'Roland Robot',
                    'address_line1': '42 Sevenoaks St',
                    'address_line2': None,
                    'address_city': 'Lathlain',
                    'address_postcode': '6454',
                    'address_state': 'WA',
                    'address_country': 'Australia',
                    'primary': None,
                },
                'transfer': None
            }
        })
        self.response_error = json.dumps({
            'error': 'invalid_resource',
            'error_description':
                'One or more parameters were missing or invalid.',
            # Should there really be a charge token?
            'charge_token': '1234',
            'messages': [{
                'code': 'description_invalid',
                'message': 'Description can\'t be blank',
                'param': 'description'
            }]
        })
        self.response_error_no_messages = json.dumps({
            'error': 'invalid_resource',
            'error_description':
                'One or more parameters were missing or invalid.',
            # Should there really be a charge token?
            'charge_token': '1234'
        })

    @patch('requests.post')
    def test_only_process_once(self, mock_request):
        """ Check that transactions are processed exactly once """
        mock_request.return_value = FakeResponse(200, self.response_data)

        # Shouldn't be marked as processed before process_transaction is called
        # for the first time.
        self.assertFalse(self.transaction.processed)

        # Should be marked after the first call.
        result = self.transaction.process_transaction()
        self.assertTrue(self.transaction.processed)

        # Shouldn't process anything the second time
        result = self.transaction.process_transaction()
        self.assertIsNone(result)

    @override_settings(PIN_ENVIRONMENTS={})
    @patch('requests.post')
    def test_valid_environment(self, mock_request):
        """ Check that an error is thrown with no environment """
        mock_request.return_value = FakeResponse(200, self.response_data)
        self.assertRaises(PinError, self.transaction.process_transaction)

    @override_settings(PIN_ENVIRONMENTS=ENV_MISSING_SECRET)
    @patch('requests.post')
    def test_secret_set(self, mock_request):
        """ Check that an error is thrown with no secret """
        mock_request.return_value = FakeResponse(200, self.response_data)
        self.assertRaises(ConfigError, self.transaction.process_transaction)

    @override_settings(PIN_ENVIRONMENTS=ENV_MISSING_HOST)
    @patch('requests.post')
    def test_host_set(self, mock_request):
        """ Check that an error is thrown with no host """
        mock_request.return_value = FakeResponse(200, self.response_data)
        self.assertRaises(ConfigError, self.transaction.process_transaction)

    @patch('requests.post')
    def test_response_not_json(self, mock_request):
        """ Check that failure is returned for non-JSON responses """
        mock_request.return_value = FakeResponse(200, '')
        response = self.transaction.process_transaction()
        self.assertEqual(response, 'Failure.')

    @patch('requests.post')
    def test_response_badparam(self, mock_request):
        """ Check that a specific error is thrown for invalid parameters """
        mock_request.return_value = FakeResponse(200, self.response_error)
        response = self.transaction.process_transaction()
        self.assertEqual(response, 'Failure: Description can\'t be blank')

    @patch('requests.post')
    def test_response_noparam(self, mock_request):
        """ Check that a specific error is thrown for missing parameters """
        mock_request.return_value = FakeResponse(
            200, self.response_error_no_messages
        )
        response = self.transaction.process_transaction()
        self.assertEqual(
            response,
            'Failure: One or more parameters were missing or invalid.'
        )

    @patch('requests.post')
    def test_response_success(self, mock_request):
        """ Check that the success response is correctly processed """
        mock_request.return_value = FakeResponse(200, self.response_data)
        response = self.transaction.process_transaction()
        self.assertEqual(response, 'Success!')
        self.assertTrue(self.transaction.succeeded)
        self.assertEqual(self.transaction.transaction_token, '12345')
        self.assertEqual(self.transaction.fees, 5.0)
        self.assertEqual(self.transaction.pin_response, 'Success!')
        self.assertEqual(self.transaction.card_address1, '42 Sevenoaks St')
        self.assertIsNone(self.transaction.card_address2)
        self.assertEqual(self.transaction.card_city, 'Lathlain')
        self.assertEqual(self.transaction.card_state, 'WA')
        self.assertEqual(self.transaction.card_postcode, '6454')
        self.assertEqual(self.transaction.card_country, 'Australia')
        self.assertEqual(self.transaction.card_number, 'XXXX-XXXX-XXXX-0000')
        self.assertEqual(self.transaction.card_type, 'master')



""" Template Tag test classes """
from datetime import datetime
from django.template import TemplateSyntaxError
from django.template.context import RequestContext
from django.test import TestCase, RequestFactory
from django.test.utils import override_settings
from pinpayments.templatetags.pin_payment_tags import pin_form, pin_header

ENV_OK = {
    'staging': {
        'key': 'key1',
        'secret': 'secret1',
        'host': 'test-api.pin.net.au',
    },
}

ENV_MISSING_KEY = {
    'test': {
        'secret': 'secret1',
        'host': 'test-api.pin.net.au',
    },
}

ENV_MISSING_HOST = {
    'test': {
        'key': 'key1',
        'secret': 'secret1',
    },
}


class TemplateTagsTests(TestCase):
    """ Test the template tags """
    def __init__(self, *args, **kwargs):
        """ define attributes """
        self.factory = None
        self.request = None
        self.context = None
        super(TemplateTagsTests, self).__init__(*args, **kwargs)

    def set_up(self):
        """ Common setup for all methods """
        self.factory = RequestFactory()
        self.request = self.factory.get('/')
        self.context = RequestContext(self.request)

    def test_pin_form(self):
        """ Check the years which are output """
        form = pin_form(self.context)
        current_year = datetime.now().year
        self.assertEqual(
            form['pin_cc_years'],
            range(current_year, current_year + 15)
        )

    def test_pin_header_default_env(self):
        """ Check 'test' is the default environment """
        header = pin_header(self.context)
        self.assertEqual(header['pin_environment'], 'test')

    @override_settings(PIN_ENVIRONMENTS={})
    def test_pin_header_no_env(self):
        """ Check errors are raised when no environments are present """
        self.assertRaises(TemplateSyntaxError, pin_header, self.context)

    def test_pin_header_bad_env(self):
        """ Check an error is raised when a bad environment is requested """
        with self.assertRaises(TemplateSyntaxError):
            pin_header(self.context, environment='should not exist')

    @override_settings(PIN_ENVIRONMENTS=ENV_MISSING_KEY)
    def test_pin_header_no_key(self):
        self.assertRaises(TemplateSyntaxError, pin_header, self.context)

    @override_settings(PIN_ENVIRONMENTS=ENV_MISSING_HOST)
    def test_pin_header_no_host(self):
        self.assertRaises(TemplateSyntaxError, pin_header, self.context)

    @override_settings(PIN_ENVIRONMENTS=ENV_OK)
    def test_pin_header_success(self):
        """ Check the values of the environment are properly extracted """
        header = pin_header(self.context, environment='staging')
        self.assertEqual(header['pin_environment'], 'staging')
        self.assertEqual(header['pin_public_key'], 'key1')
        self.assertEqual(header['pin_host'], 'test-api.pin.net.au')



from django.core.management.base import BaseCommand, CommandError
from pinpayments.models import PinPlan
from django.conf import settings

class Command(BaseCommand):
    help = 'Retrieves Plan configuration from Pin Payments. Updates all configured environments unless an optional environment name is provided.'

    def add_arguments(self, parser):
        parser.add_argument('environment', nargs='?', type=str, default='all_environments')

    def handle(self, *args, **options):
        env = options['environment']
        if env == 'all_environments':
            for pin_env in settings.PIN_ENVIRONMENTS.keys():
                if settings.PIN_ENVIRONMENTS[pin_env].get('secret'):
                    sync_env(pin_env)
        else:
            sync_env(env)

def sync_env(env):
    result = PinPlan.get_from_pin(env)
    print("[{}] Created {} plan(s)".format(env, result['new']))
    print("[{}] Updated {} plan(s)".format(env, result['updated']))



"""
Models for interacting with Pin, and storing results
"""

from datetime import datetime
from decimal import Decimal

from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.timezone import get_default_timezone
from django.utils.translation import ugettext_lazy as _

from .exceptions import ConfigError, PinError
from .objects import PinEnvironment
from .utils import get_value


if getattr(settings, 'PIN_ENVIRONMENTS', {}) == {}:
    raise ConfigError("PIN_ENVIRONMENTS not defined.")


TRANS_TYPE_CHOICES = (
    ('Payment', 'Payment'),
    ('Refund', 'Refund'),
)

CARD_TYPES = (
    ('master', 'Mastercard'),
    ('visa', 'Visa'),
)

PLAN_INTERVAL_CHOICES = (
    ('day', 'Day(s)'),
    ('week', 'Week(s)'),
    ('month', 'Month(s)'),
    ('year', 'Year(s)'),
)

class CustomerToken(models.Model):
    """
    A token returned by the Pin Payments Customer API.
    These can be used on a Transaction in lieu of of a Card token, and
    can be reused.
    They are linked to a User record and are typically used for recurring
    billing.
    Card token - difference is that a card can only be used once, for a transaction
    or to be converted to a Customer token. Customer tokens can be reused.
    """
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    environment = models.CharField(
        max_length=25, db_index=True, blank=True,
        help_text=_('The name of the Pin environment to use, eg test or live.')
    )
    token = models.CharField(
        _('Token'), max_length=100,
        help_text=_('Generated by Card API or Customers API')
    )
    created = models.DateTimeField(_('Created'), auto_now_add=True)
    active = models.BooleanField(_('Active'), default=True)
    card_type = models.CharField(
        _('Card Type'), max_length=20, blank=True, null=True,
        choices=CARD_TYPES, help_text=_('Determined automatically by Pin')
    )
    card_number = models.CharField(
        _('Card Number'), max_length=100, blank=True, null=True,
        help_text=_('Cleansed by Pin API')
    )
    card_name = models.CharField(
        _('Name on Card'), max_length=100, blank=True, null=True
    )

    def __str__(self):
        return "{0}".format(self.token)

    def save(self, *args, **kwargs):
        if not self.environment:
            self.environment = getattr(settings, 'PIN_DEFAULT_ENVIRONMENT', 'test')
        super(CustomerToken, self).save(*args, **kwargs)

    def new_card_token(self, card_token):
        """ Placeholder to retain name and functionality of old method """
        self.update_card(card_token)
        return True

    def update_card(self, card_token):
        """ Provide a card token to update the details for this customer """
        pin_env = PinEnvironment(self.environment)
        payload = {'card_token': card_token}
        url_tail = "/customers/{}".format(self.token)
        data = pin_env.pin_put(url_tail, payload)[1]['response']
        self.card_number = data['card']['display_number']
        self.card_type = data['card']['scheme']
        self.card_name = data['card']['name']
        self.save()

    @classmethod
    def create_from_card_token(cls, card_token, user, environment=''):
        """ Create a customer token from a card token """
        pin_env = PinEnvironment(environment)
        payload = {'email': user.email, 'card_token': card_token}
        data = pin_env.pin_post("/customers", payload)[1]['response']
        customer = CustomerToken.objects.create(
            user=user,
            token=data['token'],
            environment=environment,
            card_number=data['card']['display_number'],
            card_type=data['card']['scheme'],
            card_name=data['card']['name'],
        )
        return customer


class PinTransaction(models.Model):
    """
    PinTransaction - model to hold response data from the pin.net.au
    Charge API. Note we capture the card and/or customer token, but
    there's no FK to your own customers table. That's for you to do
    in your own code.
    """
    date = models.DateTimeField(
        _('Date'), db_index=True, help_text=_(
            'Time this transaction was put in the database. '
            'May differ from the time that PIN reports the transaction.'
        )
    )
    environment = models.CharField(
        max_length=25, db_index=True, blank=True,
        help_text=_('The name of the Pin environment to use, eg test or live.')
    )
    amount = models.DecimalField(
        _('Amount (Dollars)'), max_digits=10, decimal_places=2
    )
    fees = models.DecimalField(
        _('Transaction Fees'), max_digits=10, decimal_places=2,
        default=Decimal("0.00"), blank=True, null=True, help_text=_(
            'Fees charged to you by Pin, for this transaction, in dollars'
        )
    )
    description = models.TextField(
        _('Description'), blank=True, null=True,
        help_text=_('As provided when you initiated the transaction')
    )
    processed = models.BooleanField(
        _('Processed?'), default=False,
        help_text=_('Has this been sent to Pin yet?')
    )
    succeeded = models.BooleanField(
        _('Success?'), default=False,
        help_text=_('Was the transaction approved?')
    )
    currency = models.CharField(
        _('Currency'), max_length=100, default='AUD',
        help_text=_('Currency transaction was processed in')
    )
    transaction_token = models.CharField(
        _('Pin API Transaction Token'), max_length=100, blank=True, null=True,
        db_index=True, help_text=_('Unique ID from Pin for this transaction')
    )
    card_token = models.CharField(
        _('Pin API Card Token'), max_length=40, blank=True, null=True,
        help_text=_(
            'Card token used for this transaction (Card API and Web Forms)'
        )
    )
    customer_token = models.ForeignKey(
        CustomerToken, blank=True, null=True,
        help_text=_('Provided by Customer API'),
        on_delete=models.SET_NULL,
    )
    pin_response = models.CharField(
        _('API Response'), max_length=255, blank=True, null=True,
        help_text=_('Response text, usually Success!')
    )
    ip_address = models.GenericIPAddressField(
        help_text=_('IP Address used for payment')
    )
    email_address = models.EmailField(
        _('E-Mail Address'), max_length=100, help_text=_('As passed to Pin.')
    )
    card_address1 = models.CharField(
        _('Cardholder Street Address'), max_length=100, blank=True, null=True,
        help_text=_('Address entered by customer to process this transaction')
    )
    card_address2 = models.CharField(
        _('Cardholder Street Address Line 2'), max_length=100, blank=True,
        null=True
    )
    card_city = models.CharField(
        _('Cardholder City'), max_length=100, blank=True, null=True
    )
    card_state = models.CharField(
        _('Cardholder State'), max_length=100, blank=True, null=True
    )
    card_postcode = models.CharField(
        _('Cardholder Postal / ZIP Code'), max_length=100, blank=True,
        null=True
    )
    card_country = models.CharField(
        _('Cardholder Country'), max_length=100, blank=True, null=True
    )
    card_number = models.CharField(
        _('Card Number'), max_length=100, blank=True, null=True,
        help_text=_('Cleansed by Pin API')
    )
    card_type = models.CharField(
        _('Card Type'), max_length=20, blank=True, null=True,
        choices=CARD_TYPES, help_text=_('Determined automatically by Pin')
    )
    pin_response_text = models.TextField(
        _('Complete API Response'), blank=True, null=True,
        help_text=_('The full JSON response from the Pin API')
    )

    def save(self, *args, **kwargs):
        if not (self.card_token or self.customer_token):
            raise PinError("Must provide card_token or customer_token")

        if self.card_token and self.customer_token:
            raise PinError("Can only provide card_token OR customer_token, not both")

        if not self.environment:
            self.environment = getattr(settings, 'PIN_DEFAULT_ENVIRONMENT', 'test')

        if self.environment not in getattr(settings, 'PIN_ENVIRONMENTS', {}):
            raise PinError("Pin Environment '{0}' does not exist".format(self.environment))

        if not self.date:
            now = datetime.now()
            if settings.USE_TZ:
                now = timezone.make_aware(now, get_default_timezone())
            self.date = now

        super(PinTransaction, self).save(*args, **kwargs)

    def __str__(self):
        return "{0}".format(self.id)

    class Meta:
        verbose_name = 'PIN.net.au Transaction'
        verbose_name_plural = 'PIN.net.au Transactions'
        ordering = ['-date']

    def process_transaction(self):
        """ Send the data to Pin for processing """
        if self.processed:
            return None  # can only attempt to process once.
        self.processed = True
        self.save()

        pin_env = PinEnvironment(self.environment)
        payload = {
            'email': self.email_address,
            'description': self.description,
            'amount': int(self.amount * 100),
            'currency': self.currency,
            'ip_address': self.ip_address,
        }
        if self.card_token:
            payload['card_token'] = self.card_token
        else:
            payload['customer_token'] = self.customer_token.token

        response, response_json = pin_env.pin_post('/charges', payload, True)
        self.pin_response_text = response.text

        if response_json is None:
            self.pin_response = 'Failure.'
        elif 'error' in response_json.keys():
            if 'messages' in response_json.keys():
                if 'message' in response_json['messages'][0].keys():
                    self.pin_response = 'Failure: {0}'.format(
                        response_json['messages'][0]['message']
                    )
            else:
                self.pin_response = 'Failure: {0}'.format(
                    response_json['error_description']
                )
            self.transaction_token = response_json.get('charge_token', None)
        else:
            data = response_json['response']
            self.succeeded = True
            self.transaction_token = data['token']
            self.fees = data['total_fees'] / Decimal("100.00")
            self.pin_response = data['status_message']
            self.card_address1 = data['card']['address_line1']
            self.card_address2 = data['card']['address_line2']
            self.card_city = data['card']['address_city']
            self.card_state = data['card']['address_state']
            self.card_postcode = data['card']['address_postcode']
            self.card_country = data['card']['address_country']
            self.card_number = data['card']['display_number']
            self.card_type = data['card']['scheme']

        self.save()
        return self.pin_response


class BankAccount(models.Model):
    """ A representation of a bank account, as stored by Pin. """
    token = models.CharField(
        _('Pin API Bank account token'), max_length=40, db_index=True,
        help_text=_("A bank account token provided by Pin")
    )
    bank_name = models.CharField(
        _('Bank Name'), max_length=100,
        help_text=_("The name of the bank at which this account is held")
    )
    branch = models.CharField(
        _('Branch name'), max_length=100, blank=True,
        help_text=_("The name of the branch at which this account is held")
    )
    name = models.CharField(
        _('Recipient Name'), max_length=100,
        help_text="The name of the bank account"
    )
    bsb = models.IntegerField(
        _('BSB'),
        help_text=_("The BSB (Bank State Branch) code of the bank account.")
    )
    number = models.CharField(
        _('BSB'), max_length=20,
        help_text=_("The account number of the bank account")
    )
    environment = models.CharField(
        max_length=25, db_index=True, blank=True,
        help_text=_('The name of the Pin environment to use, eg test or live.')
    )

    def __str__(self):
        return "{0}".format(self.token)


class PinRecipient(models.Model):
    """
    A recipient stored for the purpose of having funds transferred to them
    """
    token = models.CharField(
        max_length=40, db_index=True,
        help_text=_("A recipient token provided by Pin")
    )
    email = models.EmailField(max_length=100, help_text=_('As passed to Pin.'))
    name = models.CharField(
        max_length=100, blank=True, null=True,
        help_text=_("Optional. The name by which the recipient is referenced")
    )
    created = models.DateTimeField(_("Time created"), auto_now_add=True)
    bank_account = models.ForeignKey(
        BankAccount, blank=True, null=True,
        on_delete=models.SET_NULL,
    )
    environment = models.CharField(
        max_length=25, db_index=True, blank=True,
        help_text=_('The name of the Pin environment to use, eg test or live.')
    )

    def __str__(self):
        return "{0}".format(self.token)

    @classmethod
    def create_with_bank_account(cls, email, account_name, bsb, number, name=""):
        """ Creates a new recipient from a provided bank account's details """
        pin_env = PinEnvironment()
        payload = {
            'email': email,
            'name': name,
            'bank_account[name]': account_name,
            'bank_account[bsb]': bsb,
            'bank_account[number]': number
        }
        data = pin_env.pin_post('/recipients', payload)[1]['response']
        bank_account = BankAccount.objects.create(
            bank_name=data['bank_account']['bank_name'],
            branch=data['bank_account']['branch'],
            bsb=data['bank_account']['bsb'],
            name=data['bank_account']['name'],
            number=data['bank_account']['number'],
            token=data['bank_account']['token'],
            environment=pin_env.name,
        )
        new_recipient = cls.objects.create(
            token=data['token'],
            email=data['email'],
            name=data['name'],
            bank_account=bank_account,
            environment=pin_env.name,
        )
        return new_recipient


class PinTransfer(models.Model):
    """
    A transfer from a PinEnvironment to a PinRecipient
    """
    transfer_token = models.CharField(
        _('Pin API Transfer Token'), max_length=100, blank=True, null=True,
        db_index=True, help_text=_('Unique ID from Pin for this transfer')
    )
    status = models.CharField(
        max_length=100, blank=True, null=True,
        help_text=_("Status of transfer at time of saving")
    )
    currency = models.CharField(
        max_length=10, help_text=_("currency of transfer")
    )
    description = models.CharField(
        max_length=100, blank=True, null=True,
        help_text=_("Description as shown on statement")
    )
    amount = models.IntegerField(help_text=_(
        "Transfer amount, in the base unit of the "
        "currency (e.g.: cents for AUD, yen for JPY)"
    ))
    recipient = models.ForeignKey(PinRecipient, blank=True, null=True, on_delete=models.SET_NULL)
    created = models.DateTimeField(auto_now_add=True)
    pin_response_text = models.TextField(
        _('Complete API Response'), blank=True, null=True,
        help_text=_('The full JSON response from the Pin API')
    )

    def __str__(self):
        return "{0}".format(self.transfer_token)

    @property
    def value(self):
        """
        Returns the value of the transfer in the representation of the
        currency it is in, without symbols
        That is, 1000 cents as 10.00, 1000 yen as 1000
        """
        return get_value(self.amount, self.currency)

    @classmethod
    def send_new(cls, amount, description, recipient, currency="AUD"):
        """ Creates a transfer by sending it to Pin """
        pin_env = PinEnvironment()
        payload = {
            'amount': amount,
            'description': description,
            'recipient': recipient.token,
            'currency': currency,
        }
        response, response_json = pin_env.pin_post('/transfers', payload)
        data = response_json['response']
        new_transfer = PinTransfer.objects.create(
            transfer_token=data['token'],
            status=data['status'],
            currency=data['currency'],
            description=data['description'],
            amount=data['amount'],
            recipient=recipient,
            pin_response_text=response.text,
        )
        return new_transfer


class PinPlan(models.Model):
    """
    A Plan from a PinEnvironment to configure payment plans, which are
    later used by Customers to get Subscriptions.

    These are kept in sync with Pin Payments via the relevant command-line
    management command and via webhooks.
    """
    environment = models.CharField(
        max_length=25, db_index=True, blank=True,
        help_text=_('The name of the Pin environment to use, eg test or live.')
    )
    plan_token = models.CharField(
        _("Pin API Plan Token"), max_length=100, blank=True, null=True,
        db_index=True, help_text=_("Unique ID from Pin for this plan")
    )
    active = models.BooleanField(
        blank=True, default=True,
        help_text=_("Is this Plan still visible on the Pin API?")
    )
    name = models.CharField(
        max_length=100, blank=True, null=True,
        help_text=_("Description as shown on statement")
    )
    currency = models.CharField(
        max_length=10, help_text=_("Currency of plan")
    )
    amount = models.IntegerField(
        help_text=_("Charge amount, in the base unit of the currency (e.g.: cents for AUD, yen for JPY)"),
    )
    interval = models.IntegerField(
        help_text=_("The interval between subsequent charges, interpreted in units defined in interval_units")
    )
    interval_unit = models.CharField(
        max_length=10, choices=PLAN_INTERVAL_CHOICES,
        help_text=_("The unit of measure applied to the interval amount."),
    )
    intervals = models.IntegerField(
        default=0, blank=True,
        help_text=_("Number of intervals before a subscription is automatically cancelled. Default 0 (no limit)"),
    )
    setup_amount = models.IntegerField(
        default=0, blank=True,
        help_text=_("Amount to charge (in the currency base unit, eg cents for AUD) at the start of the first full interval."),
    )
    trial_amount = models.IntegerField(
        default=0, blank=True,
        help_text=_("Amount the customer will be charged in the currency base unit upon initiating a trial of this plan."),
    )
    trial_interval = models.IntegerField(
        default=0, blank=True,
        help_text=_("The interval between the start of the trial period and beginning of the paid subscription proper.")
    )
    trial_interval_unit = models.CharField(
        max_length=10, choices=PLAN_INTERVAL_CHOICES,
        help_text=_("The unit of measure applied to the trial interval amount."),
    )
    permission_cancel = models.BooleanField(
        blank=True, default=True,
        help_text=_("Can the customer cancel the subscription themselves via a link in their receipts?"),
    )
    created = models.DateTimeField(auto_now_add=True)
    pin_response_text = models.TextField(
        _("Complete API Response"), blank=True, null=True,
        help_text=_("The full JSON response from the Pin API")
    )

    def __str__(self):
        return "{0}".format(self.plan_token)

    @property
    def value(self):
        """
        Returns the value of the plan in the representation of the
        currency it is in, without symbols
        That is, 1000 cents as 10.00, 1000 yen as 1000
        """
        return get_value(self.amount, self.currency)

    def save(self, *args, **kwargs):
        if not self.environment:
            self.environment = getattr(settings, 'PIN_DEFAULT_ENVIRONMENT', 'test')
        super().save(*args, **kwargs)

    @classmethod
    def get_from_pin(cls, environment=''):
        """
        Retrieves all Plans from Pin Payments API and creates local PinPlan records
        for each one. Any local PinPlan record that isn't returned by the API will
        have the status changed to 'inactive'.
        """

        new_plans = 0
        updated_plans = 0

        pin_env = PinEnvironment(environment)
        response, response_json = pin_env.pin_get('/plans', {})
        data = response_json['response']

        for plan in data:
            new_plan, created = PinPlan.objects.update_or_create(
                    plan_token=plan['token'], defaults={
                        'plan_token': plan['token'],
                        'name': plan['name'],
                        'environment': environment,
                        'amount': plan['amount'],
                        'currency': plan['currency'],
                        'setup_amount': plan['setup_amount'],
                        'trial_amount': plan['trial_amount'],
                        'interval': plan['interval'],
                        'interval_unit': plan['interval_unit'],
                        'intervals': plan['intervals'],
                        'trial_interval': plan['trial_interval'],
                        'trial_interval_unit': plan['trial_interval_unit'],
                        'permission_cancel': 'cancel' in plan['customer_permissions'],
                        'pin_response_text': response.text,
                    })
            if created:
                new_plans += 1
            else:
                updated_plans += 1

        return {'new': new_plans, 'updated': updated_plans}


class Subscription(models.Model):
    """
    An instance of a PinPlan in a PinEnvironment being paid on a regular
    basis by a CustomerToken.

    Plans define the payment rules, and are generally setup once in advance.

    Subscriptions tie a token to a plan with a set of parameters on file regarding
    trial dates etc.

    The master source of subscription data is Pin Payments and they bill
    automatically at their end based on THEIR holding of Subscription data. As
    such, we can create new Subscriptions in code but need to ensure that data
    is kept up to date via webhooks.
    """
    subscription_token = models.CharField(
        _("Pin API Subscription Token"), max_length=100, blank=True, null=True,
        db_index=True, help_text=_("Unique ID from Pin for this subscription")
    )
    environment = models.CharField(
        max_length=25, db_index=True, blank=True,
        help_text=_('The name of the Pin environment to use, eg test or live.')
    )
    plan = models.ForeignKey(PinPlan, on_delete=models.CASCADE)
    customer = models.ForeignKey(CustomerToken, on_delete=models.CASCADE)
    active = models.BooleanField(
        blank=True, default=True,
        help_text=_("Is this subscription still visible on the Pin API?")
    )
    next_billing_date = models.DateTimeField(
        blank=True, null=True,
        help_text=_("The next time this subscription will be charged"),
        )
    active_interval_started_at = models.DateTimeField(
        blank=True, null=True,
        help_text=_("When did the current trial or billing period begin?"),
        )
    active_interval_finishes_at = models.DateTimeField(
        blank=True, null=True,
        help_text=_("When will the current trial or billing period finish?"),
        )
    cancelled_at = models.DateTimeField(
        blank=True, null=True,
        help_text=_("When was the subscription cancelled?"),
        )

    def __str__(self):
        return self.plan_token

    def save(self, *args, **kwargs):
        if not self.environment:
            self.environment = getattr(settings, 'PIN_DEFAULT_ENVIRONMENT', 'test')
        super().save(*args, **kwargs)



"""
Utility functions without objects
"""
from decimal import Decimal


CURRENCIES = (
    "AUD",
    "USD",
    "NZD",
    "SGD",
    "EUR",
    "GBP",
    "CAD",
    "HKD",
    "JPY",
    "MYR",
    "THB",
    "PHP",
    "ZAR",
    "IDR",
    "TWD",
)


DECIMAL_CURRENCIES = (
    "AUD",
    "USD",
    "NZD",
    "SGD",
    "EUR",
    "GBP",
    "CAD",
    "HKD",
)


SINGLE_UNIT_CURRENCIES = (
    "JPY"
    "MYR",
    "THB",
    "PHP",
    "ZAR",
    "IDR",
    "TWD",
)

CURRENCY_DETAIL = {
    "AUD": {
        "symbol": "$",
        "name": "Australian dollar",
        },
    "USD": {
        "symbol": "$",
        "name": "US dollar",
        },
    "NZD": {
        "symbol": "$",
        "name": "New Zealand dollar",
        },
    "SGD": {
        "symbol": "$",
        "name": "Singaporean dollar",
        },
    "EUR": {
        "symbol": "€",
        "name": "Euro",
        },
    "GBP": {
        "symbol": "£",
        "name": "Pound sterling",
        },
    "CAD": {
        "symbol": "$",
        "name": "Canadian dollar",
        },
    "HKD": {
        "symbol": "$",
        "name": "Hong Kong dollar",
        },
    "JPY": {
        "symbol": "¥",
        "name": "Japanese yen",
        },
    "MYR": {
        "symbol": "RM",
        "name": "Malaysian ringgit",
        },
    "THB": {
        "symbol": "฿",
        "name": "Thai baht",
        },
    "PHP": {
        "symbol": "₱",
        "name": "Philippine peso",
        },
    "ZAR": {
        "symbol": "R",
        "name": "South African rand",
        },
    "IDR": {
        "symbol": "Rp",
        "name": "Indonesian rupiah",
        },
    "TWD": {
        "symbol": "NT$",
        "name": "New Taiwan dollar",
        },
}


def get_value(amount, currency):
    """
    Returns the value of the transfer in the representation of the
    currency it is in, without symbols
    That is, 1000 cents as 10.00, 1000 yen as 1000
    """
    if currency in DECIMAL_CURRENCIES:
        return Decimal(amount)/Decimal(100)
    return Decimal(amount)
